// TdtLabel v1.00
// for Delphi 2, 3, 4, and C++Builder 1, 3, 4
// (C) 1999, Tamas Demjen
// Email: demjen@usa.net
// Tamas Demjen on the Web: http://members.xoom.com/demjen
// This component is FREEWARE.

// Thanks to Brad Stowers (bstowers@pobox.com) for the DFS.inc file.

// The component consits of the following files:
// dtLabel.pas, dtLabel.dcr, dtLabel.txt, DFS.inc
// dtLabelEditor.pas, dtLabelEditor.dfm, dtLabelPreview.pas, dtLabelPreview.dfm
(*
   Special control characters in the Caption property:

   &l(href)      begin hyperlink
   &l()          end hyperlink
   &c(clXXX)     change color
   &c(int)       change color
   &u(0|1)       underline
   &b(0|1)       bold
   &i(0|1)       italics
   &*            non-breaking space
   &&            &
   &f(name)      change font name
   &s(size)      change font size
   &p(int,al)    new paragraph, int pixels at the end, alignment

   Example:
   Label.Caption :=
    '&b(1)TdtLabel&b(0) is downloadable from ' +
    '&l(http://members.xoom.com/demjen/builder)my site&l().';
*)

unit dtLabel;

interface

{$I DFS.INC}

{$IFDEF DFS_CPPB_3_UP}
{$ObjExportAll On}
{$R 'dtLabel.dcr'}
{$ENDIF}

uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
   StdCtrls, ExtCtrls, ShellApi;

type
{$IFDEF DFS_COMPILER_2}
   TTextLayout = (tlTop, tlCenter, tlBottom);
{$ENDIF}
   TExtendedAlignment = (eaLeft, eaRight, eaCenter, eaJustify);
   TLinkRectEvent = procedure (Rect: TRect; const HRef: String) of object;
   TLinkClickEvent = procedure (Sender: TObject; Index: Integer;
      var HRef: String; var LaunchBrowser: Boolean) of object;
   TLinkEvent = procedure (Sender: TObject; Index: Integer;
      const HRef: String) of object;
   TLink = class
   public
      HRef: String;
      Left: Integer;
      Top: Integer;
      Right: Integer;
      Bottom: Integer;
   end;
   TdtLabel = class(TCustomLabel)
   private
      FAlignment: TExtendedAlignment;
      FMargin: Integer;
      FTransparent: Boolean;
      FLayout: TTextLayout;
      FBevelInner: TPanelBevel;
      FBevelOuter: TPanelBevel;
      FBevelWidth: TBevelWidth;
      FBorderStyle: TBorderStyle;
      FOnMouseEnter: TNotifyEvent;
      FOnMouseLeave: TNotifyEvent;
      FOnLinkClick: TLinkClickEvent;
      FOnLinkDblClick: TLinkClickEvent;
      FOnLinkEnter: TLinkEvent;
      FOnLinkLeave: TLinkEvent;
      LinkList: TList;
      MouseOnLink: Integer;

      procedure WmMouseMove(var Message: TWMMouse); message WM_MOUSEMOVE;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure SetAlignment(Value: TExtendedAlignment);
      procedure SetMargin(Value: Integer);
      procedure SetLayout(Value: TTextLayout);
      procedure SetBevelInner(Value: TPanelBevel);
      procedure SetBevelOuter(Value: TPanelBevel);
      procedure SetBevelWidth(Value: TBevelWidth);
      procedure SetBorderStyle(Value: TBorderStyle);
   protected
      procedure ClearLinks;
      procedure LinkRect(Rect: TRect; const HRef: String);
      procedure Paint; override;
      procedure Click; override;
      procedure DblClick; override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
   published
      property Alignment: TExtendedAlignment
         read FAlignment write SetAlignment default eaLeft;
      property Margin: Integer
         read FMargin write SetMargin default 2;
      property Layout: TTextLayout
         read FLayout write SetLayout default tlTop;
      property BevelInner: TPanelBevel
         read FBevelInner write SetBevelInner default bvNone;
      property BevelOuter: TPanelBevel
         read FBevelOuter write SetBevelOuter default bvNone;
      property BevelWidth: TBevelWidth
         read FBevelWidth write SetBevelWidth default 1;
      property BorderStyle: TBorderStyle
         read FBorderStyle write SetBorderStyle default bsNone;
      property OnMouseEnter: TNotifyEvent
         read FOnMouseEnter write FOnMouseEnter
         {$IFDEF DFS_COMPILER_3_UP}default nil{$ENDIF};
      property OnMouseLeave: TNotifyEvent
         read FOnMouseLeave write FOnMouseLeave
         {$IFDEF DFS_COMPILER_3_UP}default nil{$ENDIF};
      property OnLinkClick: TLinkClickEvent
         read FOnLinkClick write FOnLinkClick
         {$IFDEF DFS_COMPILER_3_UP}default nil{$ENDIF};
      property OnLinkDblClick: TLinkClickEvent
         read FOnLinkDblClick write FOnLinkDblClick
         {$IFDEF DFS_COMPILER_3_UP}default nil{$ENDIF};
      property OnLinkEnter: TLinkEvent
         read FOnLinkEnter write FOnLinkEnter
         {$IFDEF DFS_COMPILER_3_UP}default nil{$ENDIF};
      property OnLinkLeave: TLinkEvent
         read FOnLinkLeave write FOnLinkLeave
         {$IFDEF DFS_COMPILER_3_UP}default nil{$ENDIF};

   {$IFDEF DFS_COMPILER_4_UP}
      property Anchors;
      property AutoSize;
      property BiDiMode;
      property Constraints;
      property DragKind;
      property ParentBiDiMode;
      property OnCanResize;
      property OnConstrainedResize;
      property OnEndDock;
      property OnStartDock;
   {$ENDIF}
      property Align;
      property Caption;
      property Color;
      property DragCursor;
      property DragMode;
      property Enabled;
      property Font;
      property ParentColor;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ShowHint;
      property Transparent;
      property Visible;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDrag;
   end;

procedure Register;

procedure DrawDtText(Canvas: TCanvas; ARect: TRect;
   const Text: String; Alignment: TExtendedAlignment;
   Layout: TTextLayout; OnLinkRect: TLinkRectEvent);

implementation

constructor TdtLabel.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   AutoSize := False;
   FAlignment := eaLeft;
   FMargin := 2;
   FTransparent := False;
   FLayout := tlTop;
   FOnMouseEnter := nil;
   FOnMouseLeave := nil;
   FOnLinkClick := nil;
   FOnLinkDblClick := nil;
   FOnLinkEnter := nil;
   FOnLinkLeave := nil;
   LinkList := TList.Create;
   Width := 185;
   Height := 41;
   FBevelInner := bvNone;
   FBevelOuter := bvNone;
   FBevelWidth := 1;
   MouseOnLink := -1;
end;

destructor TdtLabel.Destroy;
begin
   inherited Destroy;
   ClearLinks;
   LinkList.Free;
end;

procedure TdtLabel.WmMouseMove(var Message: TWMMouse);
var
   i: Integer;
   Link: TLink;
   OldMouseOnLink: Integer;
begin
   inherited;
   OldMouseOnLink := MouseOnLink;
   for i := 0 to LinkList.Count - 1 do
   begin
      Link := LinkList[i];
      if (Message.XPos >= Link.Left) and (Message.XPos < Link.Right) and
         (Message.YPos >= Link.Top)  and (Message.YPos < Link.Bottom) then
      begin
         MouseOnLink := i;
      {$IFDEF DFS_COMPILER_2}
         Cursor := crUpArrow;
      {$ELSE}
         Cursor := crHandPoint;
      {$ENDIF}
         if Assigned(OnLinkEnter) and (MouseOnLink <> OldMouseOnLink) then
            OnLinkEnter(Self, i, Link.HRef);
         Exit;
      end;
   end;
   if Assigned(OnLinkLeave) and (MouseOnLink <> -1) then
      OnLinkLeave(Self, MouseOnLink, TLink(LinkList[MouseOnLink]).HRef);
   MouseOnLink := -1;
   Cursor := crArrow;
end;

procedure TdtLabel.Click;
var
   HRef: String;
   LaunchBrowser: Boolean;
begin
   inherited Click;
   if MouseOnLink <> -1 then
   begin
      HRef := TLink(LinkList[MouseOnLink]).HRef;
      if Assigned(OnLinkClick) then
      begin
         LaunchBrowser := True;
         OnLinkClick(Self, MouseOnLink, HRef, LaunchBrowser);
         if LaunchBrowser then
            ShellExecute(Application.Handle, 'open', PChar(HRef), nil, nil, SW_SHOWDEFAULT);
      end
      else
      begin
         ShellExecute(Application.Handle, 'open', PChar(HRef), nil, nil, SW_SHOWDEFAULT);
      end;
   end;
end;

procedure TdtLabel.DblClick;
var
   HRef: String;
   LaunchBrowser: Boolean;
begin
   inherited DblClick;
   if Assigned(OnLinkClick) and (MouseOnLink <> -1) then
   begin
      HRef := TLink(LinkList[MouseOnLink]).HRef;
      LaunchBrowser := False;
      OnLinkClick(Self, MouseOnLink, HRef, LaunchBrowser);
      if LaunchBrowser then
         ShellExecute(Application.Handle, 'open', PChar(HRef), nil, nil, SW_SHOWDEFAULT);
   end;
end;

procedure TdtLabel.CMMouseEnter(var Message: TMessage);
begin
   inherited;
   if Assigned(OnMouseEnter) then OnMouseEnter(Self);
end;

procedure TdtLabel.CMMouseLeave(var Message: TMessage);
begin
   inherited;
   if Assigned(OnLinkLeave) and (MouseOnLink <> -1) then
      OnLinkLeave(Self, MouseOnLink, TLink(LinkList[MouseOnLink]).HRef);
   MouseOnLink := -1;
   if Assigned(OnMouseLeave) then OnMouseLeave(Self);
end;

procedure TdtLabel.SetAlignment(Value: TExtendedAlignment);
begin
   if FAlignment <> Value then
   begin
      FAlignment := Value;
      Invalidate;
   end;
end;

procedure TdtLabel.SetMargin(Value: Integer);
begin
   if FMargin <> Value then
   begin
      FMargin := Value;
      Invalidate;
   end;
end;

procedure TdtLabel.SetBevelInner(Value: TPanelBevel);
begin
   if FBevelInner <> Value then
   begin
      FBevelInner := Value;
      Invalidate;
   end;
end;

procedure TdtLabel.SetBevelOuter(Value: TPanelBevel);
begin
   if FBevelOuter <> Value then
   begin
      FBevelOuter := Value;
      Invalidate;
   end;
end;

procedure TdtLabel.SetBevelWidth(Value: TBevelWidth);
begin
   if FBevelWidth <> Value then
   begin
      FBevelWidth := Value;
      Invalidate;
   end;
end;

procedure TdtLabel.SetBorderStyle(Value: TBorderStyle);
begin
   if FBorderStyle <> Value then
   begin
      FBorderStyle := Value;
      Invalidate;
   end;
end;

procedure TdtLabel.SetLayout(Value: TTextLayout);
begin
   if FLayout <> Value then
   begin
      FLayout := Value;
      Invalidate;
   end;
end;

procedure TdtLabel.ClearLinks;
begin
   while LinkList.Count > 0 do
   begin
      TLink(LinkList[0]).Free;
      LinkList.Delete(0);
   end;
end;

procedure TdtLabel.LinkRect(Rect: TRect; const HRef: String);
var
   ALink: TLink;
begin
   ALink := TLink.Create;
   ALink.HRef := HRef;
   ALink.Left := Rect.Left;
   ALink.Top := Rect.Top;
   ALink.Right := Rect.Right;
   ALink.Bottom := Rect.Bottom;
   LinkList.Add(ALink);
end;

procedure TdtLabel.Paint;
var
   ARect: TRect;
   TopColor, BottomColor: TColor;

   procedure AdjustColors(Bevel: TPanelBevel);
   begin
      TopColor := clBtnHighlight;
      if Bevel = bvLowered then TopColor := clBtnShadow;
      BottomColor := clBtnShadow;
      if Bevel = bvLowered then BottomColor := clBtnHighlight;
   end;

begin
   ARect := GetClientRect;
   if BevelOuter <> bvNone then
   begin
     AdjustColors(BevelOuter);
     Frame3D(Canvas, ARect, TopColor, BottomColor, BevelWidth);
   end;
   if BorderStyle = bsSingle then
   begin
      DrawEdge(Canvas.Handle, ARect, BDR_SUNKENOUTER, BF_RECT);
      InflateRect(ARect, -1, -1);
      DrawEdge(Canvas.Handle, ARect, BDR_SUNKENINNER, BF_RECT);
      InflateRect(ARect, -1, -1);
   end;
   if BevelInner <> bvNone then
   begin
     AdjustColors(BevelInner);
     Frame3D(Canvas, ARect, TopColor, BottomColor, BevelWidth);
   end;
   with Canvas do
   begin
      if not Transparent then
      begin
         Brush.Color := Color;
         FillRect(ARect);
      end;
      Brush.Style := bsClear;
      Font := Self.Font;
      InflateRect(ARect, -Margin, -Margin);
      ClearLinks;
      DrawDtText(Canvas, ARect, Caption, Alignment, Layout, LinkRect);
   end;
end;

procedure DrawDtText(Canvas: TCanvas; ARect: TRect;
   const Text: String; Alignment: TExtendedAlignment;
   Layout: TTextLayout; OnLinkRect: TLinkRectEvent);
const
   NumColors = 16;
   ColorNames: array [1..NumColors] of String = (
      'clBlack', 'clMaroon', 'clGreen', 'clOlive', 'clNavy', 'clPurple',
      'clTeal', 'clGray', 'clSilver', 'clRed', 'clLime', 'clYellow', 'clBlue',
      'clFuchsia', 'clAqua', 'clWhite' );
   Colors: array [1..NumColors] of TColor = (
      clBlack, clMaroon, clGreen, clOlive, clNavy, clPurple,
      clTeal, clGray, clSilver, clRed, clLime, clYellow, clBlue,
      clFuchsia, clAqua, clWhite );
var
   Height, Width, TextW, tw, LineHeight, LineAscent, LineDescent: Integer;
   col, StartCol, WordCount: Integer;
   s: String;
   x, y: Integer;
   FontBackup: TFont;
   OldColor: Integer;
   OldUnderline: Boolean;
   Link, LinkBackup: Boolean;
   HRef: String;
   FirstWord: Boolean;
   RequestPara: Boolean;
   SpaceAfterPara, SpaceBeforePara: Integer;
   AlignPara: TExtendedAlignment;
   Contin: Boolean;
   Save_Font: TFont;

   procedure Parse(Paint: Boolean; AdditionalSpace: Integer);
   var
      tm: TTextMetric;
      tasc, tdesc, lastcol, SpaceToInsert, WordIndex: Integer;
      CommandMode: Boolean;
      param: String;
      paramcol, paraminc, i: Integer;
      tmps: String;
      RequestSpaceBefore, RequestSpaceAfter, SpaceAddedBefore: Boolean;
      LastCharSpace, PrevWordUnderlined, ThisWordUnderlined: Boolean;
      RequestBoldOn, RequestBoldOff: Boolean;
      RequestUnderlineOn, RequestUnderlineOff: Boolean;
      RequestItalicOn, RequestItalicOff: Boolean;
      RequestColor, RequestFont, RequestSize, RequestLink: Boolean;
      ChangeColor, ChangeSize: Integer;
      ChangeFont, ChangeLink: String;
      StartLinkX: Integer;
//    RestoreItalic: Boolean;
      LastTw: Integer;

      RequestSavePosition: Boolean;
      Save_col, Save_OldColor: Integer;
      Save_Link, Save_OldUnderline: Boolean;
      Save_HRef: String;
      Save_TextW: Integer;

      procedure PreSavePosition;
      begin
      end;

      procedure PostSavePosition;
      begin
      end;

      procedure SavePosition(lastcol: Integer);
      begin
         Save_col := col;
         Save_Font.Assign(Canvas.Font);
         Save_OldColor := OldColor;
         Save_Link := Link;
         Save_OldUnderline := OldUnderline;
         Save_HRef := HRef;
         Save_TextW := TextW;
      end;

      procedure LoadPosition;
      begin
         col := Save_col;
         Canvas.Font.Assign(Save_Font);
         OldColor := Save_OldColor;
         Link := Save_Link;
         OldUnderline := Save_OldUnderline;
         HRef := Save_HRef;
         TextW := Save_TextW;
      end;

   begin
      LastTw := 0;
      RequestSavePosition := False;
      RequestBoldOn := False;
      RequestBoldOff := False;
      RequestUnderlineOn := False;
      RequestUnderlineOff := False;
      RequestItalicOn := False;
      RequestItalicOff := False;
      RequestColor := False;
      RequestFont := False;
      RequestSize := False;
      RequestLink := False;

      StartLinkX := x;
      LastCharSpace := True;
      PrevWordUnderlined := False;
      ThisWordUnderlined := False;
      CommandMode := False;
      if not Paint then WordCount := 0;
      WordIndex := 0;
      if not Paint then
      begin
//         LineHeight := 0;
         LineAscent := 0;
         LineDescent := 0;
      end;
      FirstWord := True;
      TextW := 0;
      RequestSpaceBefore := False;
      RequestSpaceAfter := False;
      if RequestPara then
      begin
         Inc(y, SpaceBeforePara);
         RequestPara := False;
         SpaceBeforePara := SpaceAfterPara;
      end;
//      SavePosition;
      while col <= Length(Text) do
      begin
         s := '';
         lastcol := col; // no RequestSavePosition here!
         while col <= Length(Text) do
         begin
            if Text[col] = ' ' then
            begin
               Inc(col);
               lastcol := col;
               RequestSavePosition := True;
//               SavePosition;
               if s <> '' then
               begin
                  RequestSpaceAfter := True;
                  Break;
               end
               else
                  RequestSpaceBefore := True;
            end
            else if (CommandMode) and (Text[col] <> '&') then
            begin
               CommandMode := False;
               case UpCase(Text[col]) of
                  '*':
                  begin
                     s := s + ' ';
                     Inc(col);
//                     SavePosition;
                     lastcol := col;
                     RequestSavePosition := True;
                  end;
                  'B':
                  begin
                     Inc(col);
                     param := Copy(Text, col, 3);
                     Inc(col, 3);
                     if Length(param) = 3 then
                     begin
                        if (param[2] = '1') and not (fsBold in Canvas.Font.Style) then
                        begin RequestBoldOn := True; Break; end
                        else if (param[2] = '0') and (fsBold in Canvas.Font.Style) then
                        begin RequestBoldOff := True; Break; end;
                     end;
                  end;
                  'U':
                  begin
                     Inc(col);
                     param := Copy(Text, col, 3);
                     Inc(col, 3);
                     if Length(param) = 3 then
                     begin
                        if (param[2] = '1') and not (fsUnderline in Canvas.Font.Style) then
                        begin RequestUnderlineOn := True; Break; end
                        else if (param[2] = '0') and (fsUnderline in Canvas.Font.Style) then
                        begin RequestUnderlineOff := True; Break; end;
                     end;
                  end;
                  'I':
                  begin
                     Inc(col);
                     param := Copy(Text, col, 3);
                     Inc(col, 3);
                     if Length(param) = 3 then
                     begin
                        if (param[2] = '1') and not (fsItalic in Canvas.Font.Style) then
                        begin RequestItalicOn := True; Break; end
                        else if (param[2] = '0') and (fsItalic in Canvas.Font.Style) then
                        begin RequestItalicOff := True; Break; end;
                     end;
                  end;
                  'C':
                  begin
                     Inc(col);
                     param := Copy(Text, col, 100);
                     paramcol := Pos('(', param);
                     paraminc := paramcol;
                     if paramcol > 0 then
                     begin
                        Delete(param, 1, paramcol);
                        paramcol := Pos(')', param);
                        Inc(paraminc, paramcol);
                        if paramcol > 0 then
                        begin
                           Delete(param, paramcol, 100);
                           param := Trim(param);
                           try
                              ChangeColor := StrToInt(param);
                              RequestColor := True;
                           except
                              on EConvertError do
                              begin
                                 for i := 1 to NumColors do
                                 begin
                                    if param = ColorNames[i] then
                                    begin
                                       ChangeColor := Colors[i];
                                       RequestColor := True;
                                       Break;
                                    end;
                                 end;
                              end;
                           end;
                        end;
                     end;
                     if RequestColor then
                     begin
                        Inc(col, paraminc);
                        Break;
                     end;
                  end;
                  'F':
                  begin
                     Inc(col);
                     param := Copy(Text, col, 100);
                     paramcol := Pos('(', param);
                     paraminc := paramcol;
                     if paramcol > 0 then
                     begin
                        Delete(param, 1, paramcol);
                        paramcol := Pos(')', param);
                        Inc(paraminc, paramcol);
                        if paramcol > 0 then
                        begin
                           Delete(param, paramcol, 100);
                           param := Trim(param);
                           ChangeFont := param;
                           RequestFont := True;
                        end;
                     end;
                     if RequestFont then
                     begin
                        Inc(col, paraminc);
                        Break;
                     end;
                  end;
                  'S':
                  begin
                     Inc(col);
                     param := Copy(Text, col, 100);
                     paramcol := Pos('(', param);
                     paraminc := paramcol;
                     if paramcol > 0 then
                     begin
                        Delete(param, 1, paramcol);
                        paramcol := Pos(')', param);
                        Inc(paraminc, paramcol);
                        if paramcol > 0 then
                        begin
                           Delete(param, paramcol, 100);
                           param := Trim(param);
                           try
                              ChangeSize := StrToInt(param);
                              RequestSize := True;
                           except
                              on EConvertError do ;
                           end;
                        end;
                     end;
                     if RequestSize then
                     begin
                        Inc(col, paraminc);
                        Break;
                     end;
                  end;
                  'L':
                  begin
                     Inc(col);
                     param := Copy(Text, col, 100);
                     paramcol := Pos('(', param);
                     paraminc := paramcol;
                     if paramcol > 0 then
                     begin
                        Delete(param, 1, paramcol);
                        paramcol := Pos(')', param);
                        Inc(paraminc, paramcol);
                        if paramcol > 0 then
                        begin
                           Delete(param, paramcol, 100);
                           param := Trim(param);
                           ChangeLink := param;
                           RequestLink := True;
                        end;
                     end;
                     if RequestLink then
                     begin
                        Inc(col, paraminc);
                        Break;
                     end;
                  end;
                  'P':
                  begin
                     Inc(col);
                     param := Copy(Text, col, 100);
                     paramcol := Pos('(', param);
                     paraminc := paramcol;
                     if paramcol > 0 then
                     begin
                        Delete(param, 1, paramcol);
                        paramcol := Pos(')', param);
                        Inc(paraminc, paramcol);
                        if paramcol > 0 then
                        begin
                           Delete(param, paramcol, 100);
                           param := Trim(param);
                           paramcol := Pos(',', param);
                           if paramcol > 0 then
                           begin
                              try
                                 SpaceAfterPara := StrToInt(Copy(param, 1, paramcol - 1));
                                 param := Trim(Copy(param, paramcol + 1, 100));
                                 if param = 'Right' then AlignPara := eaRight
                                 else if param = 'Center' then AlignPara := eaCenter
                                 else if param = 'Justify' then AlignPara := eaJustify
                                 else AlignPara := eaLeft;
                                 RequestPara := True;
                              except
                                 on EConvertError do ;
                              end;
                           end;
                        end;
                     end;
                     if RequestPara then
                     begin
                        Inc(col, paraminc);
                        Break;
                     end;
                  end;
                  else
                     Inc(col);
               end;
            end
            else if (not CommandMode) and (Text[col] = '&') then
            begin
               CommandMode := True;
               Inc(col);
            end
            else
            begin
               CommandMode := False;
               s := s + Text[col];
               Inc(col);
            end;
         end;
         if (s = '') and (RequestPara) then
         begin
            FirstWord := False;
            Break;
         end;
         if RequestSpaceBefore and not LastCharSpace and ((s = '') or (s[1] <> ' ')) then
         begin s := ' ' + s; SpaceAddedBefore := True; end
         else SpaceAddedBefore := False;
         tw := Canvas.TextWidth(s);
//         th := Canvas.TextHeight(s);
         GetTextMetrics(Canvas.Handle, tm);
         tasc := tm.tmAscent;
         tdesc := tm.tmDescent;
{         if tm.tmAscent > tasc then tasc := tm.tmAscent;
         if tm.tmDescent > tdesc then tdesc := tm.tmDescent;}
         if not Paint then
         begin
            if tasc  > LineAscent  then LineAscent  := tasc;
            if tdesc > LineDescent then LineDescent := tdesc;
         end;
         if TextW + tw < Width then
         begin
            PreSavePosition;
            if Paint then
            begin
               if (Alignment = eaJustify) and (AdditionalSpace > 0) and
                  SpaceAddedBefore and (WordCount > 1) then
               begin
                  SpaceToInsert := AdditionalSpace div (WordCount - 1);
                  if WordIndex < AdditionalSpace mod (WordCount - 1) then
                     Inc(SpaceToInsert);
                  Inc(SpaceToInsert, Canvas.TextWidth(' '));
                  Dec(tw, Canvas.TextWidth(' '));
                  Inc(TextW, Canvas.TextWidth(' '));
                  if PrevWordUnderlined then
                  begin
                     tmps := '';
                     while Canvas.TextWidth(tmps) < SpaceToInsert do tmps := tmps + ' ';
                     Canvas.TextOut(x, y + LineHeight - LineDescent + tdesc - Canvas.TextHeight(tmps), tmps);
                  end;
                  Inc(x, SpaceToInsert);
                  Delete(s, 1, 1); // delete space, because it's just inserted
               end
               else if (s <> '') and (s <> ' ') and SpaceAddedBefore then
               begin
{
                  if fsItalic in Canvas.Font.Style then
                  begin
                     Canvas.Font.Style := Canvas.Font.Style - [fsItalic];
                     RestoreItalic := True;
                  end
                  else
                     RestoreItalic := False;
}
//                ShowMessage('a');
                  Dec(tw, Canvas.TextWidth(' '));
                  Inc(TextW, Canvas.TextWidth(' '));
                  if PrevWordUnderlined then
                     Canvas.TextOut(x, y + LineHeight - LineDescent + tdesc - Canvas.TextHeight(' '), ' ');
                  Inc(x, Canvas.TextWidth(' '));
                  Delete(s, 1, 1); // delete space, because it's just inserted
{
                  if RestoreItalic then
                     Canvas.Font.Style := Canvas.Font.Style + [fsItalic];
}
               end;
               Canvas.TextOut(x, y + LineHeight - LineDescent + tdesc - Canvas.TextHeight(s), s);
               if ThisWordUnderlined then
                  PrevWordUnderlined := True;
               Inc(x, tw);
               if (FirstWord or SpaceAddedBefore or LastCharSpace) and (s <> ' ') and (s <> '') then
//               if SpaceAddedBefore or FirstWord then
                  Inc(WordIndex);
            end
            else if (FirstWord or SpaceAddedBefore or LastCharSpace) and (s <> ' ') and (s <> '') then
            begin
               FirstWord := False;
               Inc(WordCount);
            end;
{
            else if (FirstWord) and (s <> '') then
            begin
               Inc(WordCount);
               FirstWord := False;
            end;
}
            {
            else if (SpaceAddedBefore or FirstWord) then
            begin
               Inc(WordCount);
               if FirstWord then
                  FirstWord := False;
            end;
            }
//                  FirstWord := False;
//            ShowMessage('"'+s+'",'+IntToStr(tw));
            Inc(TextW, tw);
            if s <> '' then
            begin
               if s[Length(s)] = ' ' then LastCharSpace := True
               else LastCharSpace := False;
            end;
            if LastCharSpace then
            begin
{               if (s = '') or (s = ' ') then
                  Inc(LastTw, Canvas.TextWidth(' '))
               else}
                  LastTw := Canvas.TextWidth(' ');
            end
            else {if s <> '' then}
               LastTw := 0;
            if RequestLink then
            begin
               if ChangeLink = '' then
               begin
                  if Link then
                  begin
                     Link := False;
                     if Paint and Assigned(OnLinkRect) then
                        OnLinkRect(Classes.Rect(StartLinkX, y, x, y + LineHeight), HRef);
                     if not OldUnderline then
                        RequestUnderlineOff := True;
                     if OldColor <> Canvas.Font.Color then
                     begin
                        RequestColor := True;
                        ChangeColor := OldColor;
                     end;
                  end;
               end
               else
               begin
                  Link := True;
                  HRef := ChangeLink;
                  StartLinkX := x;
                  OldUnderline := fsUnderline in Canvas.Font.Style;
                  if not OldUnderline then
                     RequestUnderlineOn := True;
                  OldColor := Canvas.Font.Color;
                  if OldColor <> clBlue then
                  begin
                     RequestColor := True;
                     ChangeColor := clBlue;
                  end;
               end;
               RequestLink := False;
            end;
            if RequestSpaceAfter then
            begin
               RequestSpaceBefore := True;
               RequestSpaceAfter := False;
            end
            else
               RequestSpaceBefore := False;
            if RequestBoldOn then
            begin
               Canvas.Font.Style := Canvas.Font.Style + [fsBold];
               RequestBoldOn := False;
            end
            else if RequestBoldOff then
            begin
               Canvas.Font.Style := Canvas.Font.Style - [fsBold];
               RequestBoldOff := False;
            end;
            if RequestUnderlineOn then
            begin
               Canvas.Font.Style := Canvas.Font.Style + [fsUnderline];
               RequestUnderlineOn := False;
               ThisWordUnderlined := True;
            end
            else if RequestUnderlineOff then
            begin
               Canvas.Font.Style := Canvas.Font.Style - [fsUnderline];
               RequestUnderlineOff := False;
               ThisWordUnderlined := False;
            end;
            if RequestItalicOn then
            begin
               Canvas.Font.Style := Canvas.Font.Style + [fsItalic];
               RequestItalicOn := False;
            end
            else if RequestItalicOff then
            begin
               Canvas.Font.Style := Canvas.Font.Style - [fsItalic];
               RequestItalicOff := False;
            end;
            if RequestColor then
            begin
               Canvas.Font.Color := ChangeColor;
               RequestColor := False;
            end;
            if RequestFont then
            begin
               Canvas.Font.Name := ChangeFont;
               RequestFont := False;
            end;
            if RequestSize then
            begin
               Canvas.Font.Size := ChangeSize;
               RequestSize := False;
            end;
            if RequestPara then
            begin
            {
               SpaceBeforePara := SpaceAfterPara;
               RequestPara := False;
               RequestNewParagraph
            }
               Break;
            end;
            if RequestSavePosition then
            begin
               SavePosition(lastcol);
               RequestSavePosition := False;
            end;
            PostSavePosition;
         end
         else
         begin
            if LastTw > 0 then
               Dec(TextW, LastTw);
            Contin := True;
//            col := lastcol;
            if Link then
            begin
               if Paint and Assigned(OnLinkRect) then
                  OnLinkRect(Classes.Rect(StartLinkX, y, x, y + LineHeight), HRef);
            end;
            LoadPosition;
            Break;
         end;
      end;
      if not Paint then
         LineHeight := LineAscent + LineDescent;
//      if WordCount > 0 then
//         ShowMessage(IntToStr(WordCount));
   end;

begin
   FontBackup := TFont.Create;
   Save_Font := TFont.Create;

   Width := ARect.Right - ARect.Left;
   Height := ARect.Bottom - ARect.Top;
   y := ARect.Top;

   col := 1;
   Link := False;
   RequestPara := False;
   SpaceAfterPara := 0;
   SpaceBeforePara := 0;

   if Layout <> tlTop then
   begin
      y := 0;
      FontBackup.Assign(Canvas.Font);

      while col <= Length(Text) do
      begin
         Parse(false, 0);
         if FirstWord then
            Exit;
         Inc(y, LineHeight);
      end;

      if Layout = tlCenter then
         y := (Height - y) shr 1
      else if Layout = tlBottom then
         y := Height - y;

      Canvas.Font.Assign(FontBackup);
      col := 1;
      Link := False;
      RequestPara := False;
      SpaceAfterPara := 0;
      SpaceBeforePara := 0;
   end;

   while col <= Length(Text) do
   begin
      FontBackup.Assign(Canvas.Font);
      LinkBackup := Link;
      StartCol := col;
      Contin := False;
      Parse(false, 0);
      if FirstWord then
         Exit;
      col := StartCol;
      if RequestPara then
         RequestPara := False;
      if (not Contin) and (Alignment = eaJustify) then
         Alignment := eaLeft;
      case Alignment of
         eaLeft:    x := ARect.Left;
         eaRight:   x := Width - TextW;
         eaCenter:  x := (Width - TextW) shr 1;
         eaJustify: x := ARect.Left;
      end;
      Canvas.Font.Assign(FontBackup);
      Link := LinkBackup;
      if Alignment = eaJustify then
         Parse(true, Width - TextW)
      else
         Parse(true, 0);
      Inc(y, LineHeight);
      if RequestPara then
      begin
         Alignment := AlignPara;
      end;
   end;

   FontBackup.Free;
   Save_Font.Free;
end;

procedure Register;
begin
   RegisterComponents('Label', [TdtLabel]);
end;

end.
