unit dtLabelEditor;

interface

{$I DFS.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, dtLabel, dtLabelPreview, Buttons, DsgnIntf;

type
  TdtLabelEditorForm = class(TForm)
    ListBox1: TListBox;
    Label1: TLabel;
    AddButton: TBitBtn;
    DeleteButton: TBitBtn;
    UpButton: TBitBtn;
    DownButton: TBitBtn;
    Memo1: TMemo;
    dtLabel1: TdtLabel;
    Label2: TLabel;
    OKButton: TBitBtn;
    CancelButton: TBitBtn;
    AlignmentCombo: TComboBox;
    Label3: TLabel;
    Label4: TLabel;
    SpaceEdit: TEdit;
    RevertButton: TBitBtn;
    PreviewButton: TBitBtn;
    BoldOnButton: TSpeedButton;
    BoldOffButton: TSpeedButton;
    ItalicOnButton: TSpeedButton;
    ItalicOffButton: TSpeedButton;
    UnderlineOnButton: TSpeedButton;
    UnderlineOffButton: TSpeedButton;
    FontNameButton: TSpeedButton;
    LinkOnButton: TSpeedButton;
    LinkOffButton: TSpeedButton;
    FontSizeButton: TSpeedButton;
    FontColorButton: TSpeedButton;
    AmpersandButton: TSpeedButton;
    ColorDialog1: TColorDialog;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ListBox1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure Memo1Change(Sender: TObject);
    procedure SpaceEditExit(Sender: TObject);
    procedure AlignmentComboChange(Sender: TObject);
    procedure RevertButtonClick(Sender: TObject);
    procedure AddButtonClick(Sender: TObject);
    procedure DeleteButtonClick(Sender: TObject);
    procedure PreviewButtonClick(Sender: TObject);
    procedure UpButtonClick(Sender: TObject);
    procedure DownButtonClick(Sender: TObject);
    procedure BoldOnButtonClick(Sender: TObject);
    procedure BoldOffButtonClick(Sender: TObject);
    procedure ItalicOnButtonClick(Sender: TObject);
    procedure ItalicOffButtonClick(Sender: TObject);
    procedure UnderlineOnButtonClick(Sender: TObject);
    procedure UnderlineOffButtonClick(Sender: TObject);
    procedure LinkOnButtonClick(Sender: TObject);
    procedure LinkOffButtonClick(Sender: TObject);
    procedure FontNameButtonClick(Sender: TObject);
    procedure FontSizeButtonClick(Sender: TObject);
    procedure FontColorButtonClick(Sender: TObject);
    procedure AmpersandButtonClick(Sender: TObject);
  private
    Paragraphs: TStringList;
    Alignment: TExtendedAlignment;
    SpaceAfter: Integer;
    UndoBuffer: String;
    FComponent: TdtLabel;

    procedure SetText(const Value: String);
    function GetText: String;
    procedure Select;
  public
    property Text: String read GetText write SetText;
    property Component: TdtLabel read FComponent  write FComponent;
  end;

  TdtCaptionPropertyEditor = class(TStringProperty)
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

var
  dtLabelEditorForm: TdtLabelEditorForm;

procedure Register;

implementation

{$R *.DFM}

procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(TCaption), TdtLabel, 'Caption',
    TdtCaptionPropertyEditor);
end;

function TdtCaptionPropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paMultiSelect, paRevertable];
end;

procedure TdtCaptionPropertyEditor.Edit;
var
  dtLabelEditorForm: TdtLabelEditorForm;
begin
  dtLabelEditorForm := TdtLabelEditorForm.Create(Application);
  try
    if GetComponent(0) is TdtLabel then
       dtLabelEditorForm.Component := TdtLabel(GetComponent(0));
    dtLabelEditorForm.Text := GetValue;
    if dtLabelEditorForm.ShowModal = mrOK then
      SetValue(dtLabelEditorForm.Text);
  finally
    dtLabelEditorForm.Free;
  end;
end;

function GetFirstWords(const Text: String): String;
var
   s: String;
   i: Integer;
   command: Integer;
   PrevSpace: Boolean;
begin
   PrevSpace := False;
   s := '';
   command := 0;
   for i := 1 to Length(Text) do
   begin
      if command = 0 then
      begin
         if Text[i] = '&' then command := 1
         else if Text[i] <> ' ' then
         begin
            s := s + Text[i];
            PrevSpace := False;
         end
         else if not PrevSpace then
         begin
            s := s + ' ';
            PrevSpace := True;
         end;
      end
      else if command = 1 then
      begin
         if Text[i] = '&' then
         begin
            s := s + '&';
            command := 0;
         end
         else if Text[i] = '*' then
         begin
            if not PrevSpace then
            begin
               s := s + ' ';
               PrevSpace := True;
            end;
            command := 0;
         end
         else
            command := 2;
      end
      else if (command = 2) and (Text[i] = ')') then
         command := 0;
      if Length(s) > 20 then
         Break;
   end;
   s := Trim(s);
   if s = '' then s := '(paragraph)';
   Result := s;
end;

function StripPara(const Text: String; var SpaceAfter: Integer; var Align: TExtendedAlignment): String;
var
   p, dpclose: Integer;
   s: String;
begin
   p := Pos('&p', Text);
   if p > 0 then
   begin
      s := Copy(Text, p, 100);
      p := Pos('(', s);
      dpclose := p;
      if p > 0 then
      begin
         Delete(s, 1, p);
         p := Pos(')', s);
         Inc(dpclose, p);
         if p > 0 then
         begin
            Delete(s, p, 100);
            p := Pos(',', s);
            if p > 0 then
            begin
               try
                  SpaceAfter := StrToInt(Copy(s, 1, p - 1));
                  s := Copy(s, p + 1, 100);
                  if s = 'Right' then Align := eaRight
                  else if s = 'Center' then Align := eaCenter
                  else if s = 'Justify' then Align := eaJustify
                  else Align := eaLeft;
                  p := Pos('&p', Text);
                  s := Text;
                  Delete(s, p, dpclose);
                  Result := s;
                  Exit;
               except
                  on EConvertError do ;
               end;
            end;
         end;
      end;
   end;
   Result := Text;
end;

procedure TdtLabelEditorForm.SetText(const Value: String);
var
   Text, s, als: String;
   p: Integer;
begin
   if Component <> nil then
   begin
      case Component.Alignment of
         eaRight: als := 'Right';
         eaCenter: als := 'Center';
         eaJustify: als := 'Justify';
         else als := 'Left';
      end;
   end
   else
      als := 'Left';
   ListBox1.Items.Clear;
   Paragraphs.Clear;
   Text := Trim(Value);
   p := Pos('&p', Text);
   if (p = 0) then
   begin
      ListBox1.Items.Add(GetFirstWords(Text));
      Paragraphs.Add('&p(0,' + als + ')' + Text);
      Exit;
   end
   else if p > 1 then
   begin
      s  := Copy(Text, 1, p - 1);
      ListBox1.Items.Add(GetFirstWords(s));
      Paragraphs.Add('&p(0,' + als + ')' + s);
      Delete(Text, 1, p - 1);
   end;
   while True do
   begin
      Delete(Text, 1, 2); // &p
      p := Pos('&p', Text);
      if p > 1 then
      begin
         s := '&p' + Copy(Text, 1, p - 1);
         ListBox1.Items.Add(GetFirstWords(s));
         Paragraphs.Add(s);
         Delete(Text, 1, p - 1);
      end
      else
      begin
         s := '&p' + Text;
         ListBox1.Items.Add(GetFirstWords(s));
         Paragraphs.Add(s);
         Break;
      end;
   end;
end;

function TdtLabelEditorForm.GetText: String;
var
   s: String;
   i: Integer;
begin
   s := '';
   for i := 0 to Paragraphs.Count - 1 do
   begin
      s := s + Paragraphs[i];
   end;
   Result := s;
end;

procedure TdtLabelEditorForm.FormCreate(Sender: TObject);
begin
   Paragraphs := TStringList.Create;
   Text := '';
   Component := nil;
end;

procedure TdtLabelEditorForm.FormDestroy(Sender: TObject);
begin
   Paragraphs.Free;
   Paragraphs := nil;
end;

procedure TdtLabelEditorForm.Select;
var
   s: String;
begin
   Memo1.Lines.Clear;
   Alignment := eaLeft;
   SpaceAfter := 0;
   s := StripPara(Paragraphs.Strings[ListBox1.ItemIndex], SpaceAfter, Alignment);
   if s <> '' then
      Memo1.Lines.Add(s)
   else
      Memo1Change(Memo1);
   SpaceEdit.Text := IntToStr(SpaceAfter);
   case Alignment of
      eaRight: AlignmentCombo.ItemIndex := 1;
      eaCenter: AlignmentCombo.ItemIndex := 2;
      eaJustify: AlignmentCombo.ItemIndex := 3;
      else AlignmentCombo.ItemIndex := 0;
   end;
end;

procedure TdtLabelEditorForm.ListBox1Click(Sender: TObject);
begin
   if ListBox1.ItemIndex <> -1 then
   begin
      DeleteButton.Enabled := True;
      DownButton.Enabled := ListBox1.ItemIndex < ListBox1.Items.Count - 1;
      UpButton.Enabled := ListBox1.ItemIndex > 0;
      AlignmentCombo.Enabled := True;
      SpaceEdit.Enabled := True;
      UndoBuffer := Paragraphs.Strings[ListBox1.ItemIndex];
      Select;
   end
   else
   begin
      UpButton.Enabled := False;
      DownButton.Enabled := False;
      DeleteButton.Enabled := False;
      Memo1.Clear;
      dtLabel1.Caption := '';
      AlignmentCombo.ItemIndex := 0;
      AlignmentCombo.Enabled := False;
      SpaceEdit.Text := '';
      SpaceEdit.Enabled := False;
   end;
end;

procedure TdtLabelEditorForm.FormShow(Sender: TObject);
begin
   if ListBox1.Items.Count > 0 then
   begin
      ListBox1.ItemIndex := 0;
      ListBox1Click(ListBox1);
   end;
end;

procedure TdtLabelEditorForm.Memo1Change(Sender: TObject);
var
   s: String;
   i: Integer;
   found: Boolean;
begin
   found := False;
   for i := 0 to Memo1.Lines.Count - 1 do
   begin
      if Pos('&p', Memo1.Lines[i]) > 0 then
      begin
         found := True;
         Break;
      end;
   end;
   if not found then
   begin
      case Alignment of
         eaRight: s := 'Right';
         eaCenter: s := 'Center';
         eaJustify: s := 'Justify';
         else s := 'Left';
      end;
      s := '&p(' + IntToStr(SpaceAfter) + ',' + s + ')';
      for i := 0 to Memo1.Lines.Count - 1 do
         s := s + Memo1.Lines[i] + ' ';
      while Pos('  ', s) > 0 do Delete(s, Pos('  ', s), 1);
      if s[Length(s)] = ' ' then Delete(s, Length(s), 1);
      if ListBox1.ItemIndex <> -1 then
      begin
      	i:= ListBox1.ItemIndex;
         Paragraphs.Strings[i] := s;
         ListBox1.Items[i] := GetFirstWords(s);
		{$IFDEF DFS_COMPILER_2}
         ListBox1.ItemIndex := i; // This line of code is necessary
                                  // in C++Builder 1 and Delphi 2,
                                  // otherwise the ItemIndex becomes -1
      {$ENDIF}
      end;
      dtLabel1.Caption := s;
   end
   else
   begin
      s := Memo1.Lines[i];
      Delete(s, Pos('&p', s), 2);
      Memo1.Lines[i] := s;
   end;
end;

procedure TdtLabelEditorForm.SpaceEditExit(Sender: TObject);
begin
   try
      SpaceAfter := StrToInt(SpaceEdit.Text);
      Memo1Change(Memo1);
   except
      on EConvertError do
         SpaceEdit.Text := IntToStr(SpaceAfter);
   end;
end;

procedure TdtLabelEditorForm.AlignmentComboChange(Sender: TObject);
begin
   case AlignmentCombo.ItemIndex of
      1: Alignment := eaRight;
      2: Alignment := eaCenter;
      3: Alignment := eaJustify;
      else Alignment := eaLeft;
   end;
   Memo1Change(Memo1);
end;

procedure TdtLabelEditorForm.RevertButtonClick(Sender: TObject);
begin
   if ListBox1.ItemIndex <> -1 then
   begin
      Paragraphs.Strings[ListBox1.ItemIndex] := UndoBuffer;
      Select;
   end;
end;

procedure TdtLabelEditorForm.AddButtonClick(Sender: TObject);
begin
   ListBox1.Items.Add('');
   Paragraphs.Add('');
   ListBox1.ItemIndex := ListBox1.Items.Count - 1;
   UndoBuffer := Paragraphs.Strings[ListBox1.ItemIndex];
   ListBox1Click(ListBox1);
end;

procedure TdtLabelEditorForm.DeleteButtonClick(Sender: TObject);
var
   i: Integer;
begin
   if ListBox1.ItemIndex <> -1 then
   begin
      if MessageBox(0, 'Are you sure you want to delete the selected paragraph?',
         'Delete Paragraph', MB_YESNO or MB_ICONQUESTION) = IDYES then
      begin
         i := ListBox1.ItemIndex;
         ListBox1.ItemIndex := -1;
         ListBox1.Items.Delete(i);
         Paragraphs.Delete(i);
         Dec(i);
         if (i = -1) and (ListBox1.Items.Count > 0) then
            ListBox1.ItemIndex := 0
         else
            ListBox1.ItemIndex := i;
         ListBox1Click(ListBox1);
      end;
   end;
end;

procedure TdtLabelEditorForm.PreviewButtonClick(Sender: TObject);
var
  dtLabelPreviewForm: TdtLabelPreviewForm;
begin
  dtLabelPreviewForm := TdtLabelPreviewForm.Create(Self);
  try
    dtLabelPreviewForm.dtLabel1.Caption := Text;
    dtLabelPreviewForm.ShowModal;
  finally
    dtLabelPreviewForm.Free;
  end;
end;

procedure TdtLabelEditorForm.UpButtonClick(Sender: TObject);
var
   i: Integer;
   s: String;
begin
   i := ListBox1.ItemIndex;
   ListBox1.ItemIndex := -1;
   s := ListBox1.Items[i];
   ListBox1.Items.Delete(i);
   ListBox1.Items.Insert(i - 1, s);
   s := Paragraphs[i];
   Paragraphs.Delete(i);
   Paragraphs.Insert(i - 1, s);
   ListBox1.ItemIndex := i - 1;
   ListBox1Click(ListBox1);
end;

procedure TdtLabelEditorForm.DownButtonClick(Sender: TObject);
var
   i: Integer;
   s: String;
begin
   i := ListBox1.ItemIndex;
   ListBox1.ItemIndex := -1;
   s := ListBox1.Items[i];
   ListBox1.Items.Delete(i);
   ListBox1.Items.Insert(i + 1, s);
   s := Paragraphs[i];
   Paragraphs.Delete(i);
   Paragraphs.Insert(i + 1, s);
   ListBox1.ItemIndex := i + 1;
   ListBox1Click(ListBox1);
end;

procedure TdtLabelEditorForm.BoldOnButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
      Memo1.SelText := '&b(1)';
end;

procedure TdtLabelEditorForm.BoldOffButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
      Memo1.SelText := '&b(0)';
end;

procedure TdtLabelEditorForm.ItalicOnButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
      Memo1.SelText := '&i(1)';
end;

procedure TdtLabelEditorForm.ItalicOffButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
      Memo1.SelText := '&i(0)';
end;

procedure TdtLabelEditorForm.UnderlineOnButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
      Memo1.SelText := '&u(1)';
end;

procedure TdtLabelEditorForm.UnderlineOffButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
      Memo1.SelText := '&u(0)';
end;

procedure TdtLabelEditorForm.LinkOnButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
   begin
      Memo1.SelText := '&l(http://)';
      Memo1.SelStart := Memo1.SelStart - 1;
   end;
end;

procedure TdtLabelEditorForm.LinkOffButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
      Memo1.SelText := '&l()';
end;

procedure TdtLabelEditorForm.FontNameButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
   begin
      Memo1.SelText := '&f()';
      Memo1.SelStart := Memo1.SelStart - 1;
   end;
end;

procedure TdtLabelEditorForm.FontSizeButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
   begin
      Memo1.SelText := '&s()';
      Memo1.SelStart := Memo1.SelStart - 1;
   end;
end;

procedure TdtLabelEditorForm.FontColorButtonClick(Sender: TObject);
var
   s: String;
begin
   if Memo1.Focused then
   begin
      if ColorDialog1.Execute then
      begin
         case ColorDialog1.Color of
            clBlack: s := 'clBlack';
            clMaroon: s := 'clMaroon';
            clGreen: s := 'clGreen';
            clOlive: s := 'clOlive';
            clNavy: s := 'clNavy';
            clPurple: s := 'clPurple';
            clTeal: s := 'clTeal';
            clGray: s := 'clGray';
            clSilver: s := 'clSilver';
            clRed: s := 'clRed';
            clLime: s := 'clLime';
            clYellow: s := 'clYellow';
            clBlue: s := 'clBlue';
            clFuchsia: s := 'clFuchsia';
            clAqua: s := 'clAqua';
            clWhite: s := 'clWhite';
            else s := '$' + IntToHex(Integer(ColorDialog1.Color), 8);
         end;
         Memo1.SelText := '&c(' + s + ')';
      end;
   end;
end;

procedure TdtLabelEditorForm.AmpersandButtonClick(Sender: TObject);
begin
   if Memo1.Focused then
      Memo1.SelText := '&&';
end;

end.
